import sys
import logging

class _BraceString(str):
    def __mod__(self, other):
        return self.format(*other)
    def __str__(self):
        return self


class _StyleAdapter(logging.LoggerAdapter):
    def __init__(self, logger, extra=None):
        super(_StyleAdapter, self).__init__(logger, extra)

    def process(self, msg, kwargs):
        msg = _BraceString(msg)
        return msg, kwargs
    
def _getloggers(name):
    logger   = _StyleAdapter(logging.getLogger(name))
    debug    = logger.debug
    info     = logger.info
    warning  = logger.warning
    error    = logger.error
    critical = logger.critical
    return logger, debug, info, warning, error, critical

logging.basicConfig(stream=sys.stderr)




import struct as _struct
################################################################################
### logging levels
from logging import CRITICAL as L_CRITICAL
from logging import DEBUG    as L_DEBUG
from logging import ERROR    as L_ERROR
from logging import INFO     as L_INFO
from logging import NOTSET   as L_NOTSET
from logging import WARN     as L_WARN

################################################################################
### FVWM contexts

C_NO_CONTEXT  =     0 # no context
C_WINDOW      =     1 # an application window
C_TITLE       =     2 # a title bar
C_ICON        =     4 # an icon window
C_ROOT        =     8 # the root window
C_FRAME       =    16 # a corner piece
C_SIDEBAR     =    32 # a side-bar
C_L1          =    64 # left button #1
C_L2          =   128 # left button #2
C_L3          =   256 # left button #3
C_L4          =   512 # left button #4
C_L5          =  1024 # left button #5
C_R1          =  2048 # right button #1
C_R2          =  4096 # right button #2
C_R3          =  8192 # right button #3
C_R4          = 16384 # right button #4
C_R5          = 32768 # right button #5
C_UNKNOWN     = -1

contexts=( "C_NO_CONTEXT", "C_WINDOW", "C_TITLE", "C_ICON", "C_ROOT",
           "C_FRAME", "C_SIDEBAR",
           "C_L1", "C_L2", "C_L3", "C_L4", "C_L5", "C_R1", "C_R2", "C_R3",
           "C_R4", "C_R5", "C_UNKNOWN" )

contextnames = { globals()[k] : k for k in contexts}
contextcodes = { k : globals()[k] for k in contexts}
del contexts

################################################################################
### FVWM packet types.

### We use different conventions for extended packets
### so that fvwm's (1<<31) + (1<<n) becomes (1 << (32+n))
### Conversion is taken care of automatically when reading the packet
### from the pipe or when setting masks.
### In addition M_UNKNOWN = (1 << 37) is introduced, to assign to packets that
### are not recognized 
M_NEW_PAGE       	= (1 << 0 )
M_NEW_DESK             	= (1 << 1 )
M_OLD_ADD_WINDOW       	= (1 << 2 )
M_RAISE_WINDOW         	= (1 << 3 )
M_LOWER_WINDOW         	= (1 << 4 )
M_OLD_CONFIGURE_WINDOW 	= (1 << 5 )
M_FOCUS_CHANGE         	= (1 << 6 )
M_DESTROY_WINDOW       	= (1 << 7 )
M_ICONIFY              	= (1 << 8 )
M_DEICONIFY            	= (1 << 9 )
M_WINDOW_NAME          	= (1 << 10)
M_ICON_NAME            	= (1 << 11)
M_RES_CLASS            	= (1 << 12)
M_RES_NAME             	= (1 << 13)
M_END_WINDOWLIST       	= (1 << 14)
M_ICON_LOCATION        	= (1 << 15)
M_MAP                  	= (1 << 16)
M_ERROR                	= (1 << 17)
M_CONFIG_INFO          	= (1 << 18)
M_END_CONFIG_INFO      	= (1 << 19)
M_ICON_FILE            	= (1 << 20)
M_DEFAULTICON          	= (1 << 21)
M_STRING               	= (1 << 22)
M_MINI_ICON            	= (1 << 23)
M_WINDOWSHADE          	= (1 << 24)
M_DEWINDOWSHADE        	= (1 << 25)
M_VISIBLE_NAME         	= (1 << 26)
M_SENDCONFIG           	= (1 << 27)
M_RESTACK              	= (1 << 28)
M_ADD_WINDOW           	= (1 << 29)
M_CONFIGURE_WINDOW     	= (1 << 30)
M_EXTENDED_MSG         	= (1 << 31)

MX_VISIBLE_ICON_NAME   	= (1 << 32 )
MX_ENTER_WINDOW        	= (1 << 33 )
MX_LEAVE_WINDOW        	= (1 << 34 )
MX_PROPERTY_CHANGE      = (1 << 35 )
MX_REPLY               	= (1 << 36 )

M_UNKNOWN1              = (1 << 37 )

M_ALL = ((1<<37)-1)

### These are packets that are sent when window list is requested
M_FOR_WINLIST  = ( M_CONFIGURE_WINDOW | M_ADD_WINDOW | M_WINDOW_NAME |
                   M_ICON_NAME |
                   M_VISIBLE_NAME | MX_VISIBLE_ICON_NAME | M_RES_CLASS |
                   M_RES_NAME | M_MINI_ICON | M_END_WINDOWLIST |
                   M_DESTROY_WINDOW )

### This are packets that are sent when configuration is requested
M_FOR_CONFIG   = ( M_CONFIG_INFO | M_END_CONFIG_INFO | M_SENDCONFIG )

packets = { "M_NEW_PAGE", "M_NEW_DESK", "M_OLD_ADD_WINDOW", "M_RAISE_WINDOW",
            "M_LOWER_WINDOW", "M_OLD_CONFIGURE_WINDOW", "M_FOCUS_CHANGE",
            "M_DESTROY_WINDOW", "M_ICONIFY", "M_DEICONIFY", "M_WINDOW_NAME",
            "M_ICON_NAME", "M_RES_CLASS", "M_RES_NAME", "M_END_WINDOWLIST",
            "M_ICON_LOCATION", "M_MAP", "M_ERROR", "M_CONFIG_INFO",
            "M_END_CONFIG_INFO", "M_ICON_FILE", "M_DEFAULTICON", "M_STRING",
            "M_MINI_ICON", "M_WINDOWSHADE", "M_DEWINDOWSHADE",
            "M_VISIBLE_NAME", "M_SENDCONFIG", "M_RESTACK", "M_ADD_WINDOW",
            "M_CONFIGURE_WINDOW", "M_EXTENDED_MSG", "MX_VISIBLE_ICON_NAME",
            "MX_ENTER_WINDOW", "MX_LEAVE_WINDOW", "MX_PROPERTY_CHANGE",
            "MX_REPLY", "M_UNKNOWN1" }

### get the name of the packet from value
packetnames = { globals()[k] : k for k in packets }
packetcodes = { k : globals()[k] for k in packets }
del packets

#################################################################
### communication with FVWM
### First word in packets sent from FVWM.
FVWM_PACK_START    = int("0xffffffff",0)
FVWM_PACK_START_b  = _struct.pack("L",FVWM_PACK_START)

NOT_FINISHED       = _struct.pack("L",1)
FINISHED           = _struct.pack("L",0)

LONG_SIZE          = _struct.calcsize("L")

### This is used for encoding/decoding strings sent to/from fvwm
#FVWM_STR_CODEX             = "ascii"
FVWM_STR_CODEX     = "utf8"



import struct
import sys
import time

from   .constants  import *
from   .exceptions import *
from   .log        import  _getloggers

################################################################################
### some helpers
### convert between our and FVWM packet types 
def _ptype_f2m(t):
    """Transform packet type from FVWM representation to fvwmpy 
    representation.
    """
    if not t & M_EXTENDED_MSG:
        return t
    else:
        return ( t & (M_EXTENDED_MSG - 1) ) << 32

def _ptype_m2f(t):
    """Transform packet type from fvwmpy representation to FVWM
    representation.
    """
    if t <= M_EXTENDED_MSG:
        return t
    else:
        return (t >> 32) | M_EXTENDED_MSG

### This is like struct.unpack_from() except returning value instead of
### a list of lenth 1
def _unpack(fmt,body,offset):
    x = struct.unpack_from(fmt,body,offset)
    return x[0] if len(x) == 1 else x

################################################################################
### packet does the actual reading and parsing

class packet(dict):
    """
    Instances are packets received from fvwm. 
 
    p = FvwmPacket(buf) 

    read the packet from buf. 
    Attributes can be equivallently accessed as p.key or p["key"].
    Attributes always present:
    p.ptype -- integer representing the type of the packet.
               See Fvwm.constants.M[X]_* for the types.
    p.name  -- string representation of the type of the packet.
    p.time  -- timestamp
    p.body  -- bytearray containing the body of the packet without the header
 
    Other attributes depend on the type of the packet. 
    """
    ( logger, debug, info,
      warn,   error, critical  ) = _getloggers("fvwmpy:packet")
    logger.setLevel(L_WARN)
    
    ### standard formats for reading packet fields
    ### list of pairs (<name_of_field>, <format>)
    ### formats can be anything acceptable to struct.unpack
    ### or "string" or "raw".
    ### There is another possibility
    ### (<name_of_field>, "list", <format>)
    ### In that case the value at that key will be a list of values,
    ### corresponding to format, that are read until input is exhosted.
    ### If the name of the field is an empty string, then the
    ### corresponding value is skipped over and not stored.
    ### If the name_of_field == "", then the value is skipped over.
    ### See vpacket.h in fvwm source tree and
    ### https://www.fvwm.org/Archive/ModuleInterface/
    
    ### Common parts
    ### window id -- 3 unsigned longs (last to ignore)
    _wid     = ( ("window","L"), ("frame", "L"), ("", "L") )
    ### These we get with M_[DE]ICONIFY packs
    _iconify = ( ("ix", "L"), ("iy", "L"),
                 ("idx", "L"), ("idy", "L"),
                 ("fx", "L"), ("fy", "L"),
                 ("fdx", "L"), ("fdy", "L") )
    ### empty packet or just save the raw body
    _empty   = tuple()

    ### ignore 3 unsigned longs 
    _skip3L  = ( ("","L"), ("","L"), ("","L") )

    ### this formats are derived from vpacket.h in fvwm source tree
    ### We get this with M_(CONFIGURE|ADD)_WINDOW
    ### Consult fvwm source for the meaning of the flags
    _addwindow = ( _wid +
                   ( ( "wx", "l" ),
                     ( "wy", "l" ),
                     ( "wdx", "L" ),
                     ( "wdy", "L" ),
                     ( "desk", "L" ), 
                     ( "layer", "L" ), 
                     ( "hints_base_width", "L" ), 
                     ( "hints_base_height", "L" ), 
                     ( "hints_width_inc", "L" ), 
                     ( "hints_height_inc", "L" ), 
                     ( "orig_hints_width_inc", "L" ), 
                     ( "orig_hints_height_inc", "L" ), 
                     ( "hints_min_width", "L" ), 
                     ( "hints_min_height", "L" ), 
                     ( "hints_max_width", "L" ), 
                     ( "hints_max_height", "L" ), 
                     ( "icon_w", "L" ), 
                     ( "icon_pixmap_w", "L" ), 
                     ( "hints_win_gravity", "L" ), 
                     ( "text_pixel", "L" ), 
                     ( "back_pixel", "L" ), 
                     ( "ewmh_hint_layer", "L" ), 
                     ( "ewmh_hint_desktop", "L" ), 
                     ( "ewmh_window_type", "L" ), 
                     ( "title_height", "H" ), 
                     ( "border_width", "H" ), 
                     ( "", "H" ), ### Two short dummies 
                     ( "", "H" ),
                     ### consult fvwm source for the meaning of the flags
                     ( "flags", "raw") ) 
                    )
    
    _packetformats = {    
        M_NEW_PAGE:       ( ("px", "L"), ("py", "L"), ("desk", "L"),
                            ("max_x", "L"), ("max_y", "L"), ("nx", "L"),
                            ("ny", "L") ),
        M_NEW_DESK:       ( ("desk", "L"), ),
        M_OLD_ADD_WINDOW: _empty,
        M_RAISE_WINDOW:   _wid,
        M_LOWER_WINDOW:   _wid,
        M_OLD_CONFIGURE_WINDOW:
                          _empty,
        M_FOCUS_CHANGE:   ( ("window", "L"), ("frame", "L"),
                            ("focus_change_type", "L"), ("text_pix", "L"),
                            ("border_pix", "L") ),
        M_DESTROY_WINDOW: _wid,
        M_ICONIFY:        _wid + _iconify,
        M_DEICONIFY:      _wid + _iconify,
        M_WINDOW_NAME:    _wid + ( ("win_name", "string"), ),
        M_ICON_NAME:      _wid + ( ("ico_name", "string"), ),
        M_RES_CLASS:      _wid + ( ("res_class", "string"), ),
        M_RES_NAME:       _wid + ( ("res_name", "string"), ),
        M_END_WINDOWLIST: _empty,
        M_ICON_LOCATION:  _wid + ( ("ix", "L"), ("iy", "L"),
                                  ("idx", "L"), ("idy", "L") ),
        M_MAP:            _wid,
        M_ERROR:          _skip3L + ( ("string", "string"), ), 
        M_CONFIG_INFO:    _skip3L + ( ("string", "string"), ),
        M_END_CONFIG_INFO: _empty,
        M_ICON_FILE:      _wid + ( ("ico_filename", "string"), ),
        M_DEFAULTICON:    ( ("ico_defaultfilename", "string"), ),
        M_STRING:         _wid + ( ("string", "string"), ),
        M_MINI_ICON:      _wid + ( ("mini_ico_dx", "L"), ("mini_ico_dy", "L"),
                                  ("mini_ico_depth", "L"), ("winid_pix", "L"),
                                  ("winid_mask", "L"),
                                  ("mini_ico_filename", "string") ),
        M_WINDOWSHADE:    _wid,
        M_DEWINDOWSHADE:  _wid,
        M_VISIBLE_NAME:   _wid + ( ("win_vis_name", "string"), ),
        M_SENDCONFIG:     ( ("string", "string"), ),
        M_RESTACK:        ( ( "win_stack", "listof", "3L" ), ),
        M_ADD_WINDOW:     _addwindow,
        M_CONFIGURE_WINDOW: _addwindow,
        M_EXTENDED_MSG:     _empty,

        MX_VISIBLE_ICON_NAME:
                          _wid + ( ("ico_vis_name", "string"), ),
        MX_ENTER_WINDOW:  _wid,
        MX_LEAVE_WINDOW:  _wid,
        MX_PROPERTY_CHANGE:
                          ( ("prop_type", "L"), ("val_1", "L"), ("val_2", "L"),
                            ("prop_str", "string") ),
        MX_REPLY:         _wid + ( ("string", "string"), ),
        M_UNKNOWN1:       _empty
        }

    def __init__(self,buf):
        try:
            (start,ptype,size,time) = struct.unpack_from(
                "4L", buf.read(LONG_SIZE*4), 0 )
        except struct.error:
            raise PipeDesync("Can not read the head of the packet")
        if start != FVWM_PACK_START:
            raise PipeDesync(
                "Expected {}, got {} at the beginning of the packet".
                format(hex(FVWM_PACK_START),hex(start)) )
        
        self.ptype = _ptype_f2m(ptype)
        self.time  = time
        self.debug("Read {} at {}",packetnames[self.ptype],self.time)
        ### Read and parse the rest of the packet according to the format
        ### corresponding to the type of the packet
        body = buf.read(LONG_SIZE * (size-4))
        ### save the raw body of the packet
        self.body = body
        fmt = self._packetformats[self.ptype]
        offset = 0
        for field in fmt:
            self.debug("field",field)
            self.debug("offset",offset)
            if field[1] == "string":
                self[field[0]] = ( body[offset:].
                                   decode(errors='replace').
                                   strip("\x00") )
                offset = len(body)
                break
            elif field[1] == "raw":
                self[field[0]] = body[offset:]
                offset = len(body)
                break
            elif field[1] == "listof":
                self[field[0]] = list()
                itemsize = struct.calcsize(field[2])
                while offset + itemsize <= len(body):
                    self[field[0]].append(_unpack(field[2],body,offset)) 
                    offset += itemsize
                if offset != len(body):
                    raise PipeDesync(
                        """While parsing packet {}:
                        Format doesn't match the body of the packet 
                        (while reading 'listof' format)
                        """.format(packetnames[self.ptype]))
            else:
                try:
                    self[field[0]] = _unpack(field[1], body, offset)
                except struct.error:
                    raise PipeDesync(
                        "While parsing packet {}: ".
                        format(packetnames[self.name]) +
                        "Format doesn't match the body of the packet " +
                        "(body of the packet is too short)" )

                offset += struct.calcsize( field[1] )
            ### There are variable length packs with optional fields at
            ### the end, like M_ICONIFY, so we break if the body of the pack
            ### is exhosted.
            if offset == len(body): break
        if offset != len(body):
            ### Shall we raise en exception here???
            self.error( "While parsing packet {}: " +
                        "Body of the packet is too long",
                        packetnames[self.ptype] )
        ### remove dummies
        if "" in self: del self[""]
        
    @property
    def name(self):
        """
        Return the string representation of the type of the packet
        """
        return packetnames[self.ptype]
    
    def __getattr__(self,attr):
        return self[attr]

    def __setattr__(self,attr,val):
        self[attr] = val

    def __delattr__(self,attr):
        del self[attr]

    def __str__(self):
        res = list()
        res.append( "Fvwm Packet: {} at {}".format(self.name,self.time) )
        for k,v in self.items():
            if ( self.logger.getEffectiveLevel() <= L_ERROR and
                 k=="body"):
                continue 
            elif k=="time": continue
            elif k in {"window","frame"}:
                res.append("\t| {} = {:x}".format(k,v))
            else:
                res.append("\t| {} = {}".format(k,v))
        return "\n".join(res)




class FvwmPyException(Exception):
    """
    Base exception for fvwmpy
    """
    pass

class NoException(FvwmPyException): pass

class FvwmLaunch(FvwmPyException):
    """
    fvwm module should only be executed by fvwm!
    """
    pass

class PipeDesync(FvwmPyException):
    """ 
    Indicates that packet from Fvwm can not be read or
    parsed, probably because some pipe desyncronization.
    """
    pass

class FvwmError(FvwmPyException):
    """ 
    This exception is raised, when we were trying to get some information from
    FVWM, but got M_ERROR packet in return.
    """
    pass

class IllegalOperation(FvwmPyException):
    """ 
    Indicates that some illegal action was attempted
    like assigning to Fvwm variable.
    """
    pass



import struct
import sys
import threading
import time

from   .constants  import *
from   .exceptions import *
from   .log        import  _getloggers
from   .packet     import packet

################################################################################
### some helpers

################################################################################
###

class _packet_queue:
    """Instance of the _packet_reader class represents the packet queue."""
    
    def __init__(self,module):
        ### I have to do it here because I need module.alias
        ### _packet_reader will be instancieated once per module
        ### so it is ok.
        ( self.logger, self.debug, self.info,
          self.warn,   self.error, self.critical  ) = _getloggers(
              module.alias+':packetreader')
        self.logger.setLevel(L_WARN)
        self._pipe            = module._fromfvwm
        self._queue           = list()
        self._nonempty        = threading.Event()
        self._lock            = threading.Lock()
        self._spack_found     = threading.Event()
        self._spack_picker    = None
        self._spack           = None
        self._thread_exception = None
        self._packet_picker   = None
        self._reader_thread   = threading.Thread( target = self._reader,
                                                  name   = "reader_thread",
                                                  daemon = True            )
        self.debug(" Start reader_thread as daemon")
        self._reader_thread.start()

    def __bool__(self):
        return bool(self._queue)

    def __len__(self):
        return len(self._queue)

    ### This is the one to be threaded (daemon)
    def _reader(self):
        while True:
            try:
                p = packet(self._pipe)
                self.debug("threaded_reader: got {} at {}",p.name,p.time)
                ### for testing DON'T FORGET to remove!!!
                if p.ptype == M_STRING and p.string == "exception":
                    raise Exception
                
                self._lock.acquire()
                self._queue.append(p)
                ### Are we waiting for some special packet?
                if self._spack_picker and self._spack_picker(p):
                    ### Were waiting and packer arrived
                    self.debug("threaded_reader: found special pack {}",p.name)
                    self._spack = (len(self._queue)-1, p)
                    self._spack_picker = None
                    self._spack_found.set()
                self._nonempty.set()
            except PipeDesync as e:
                self.error( "threaded_reader: {}",repr(e))
                self.error( "threaded_reader: Resync the pipe. "+
                            "Packet(s) may be lost.")
                self._resync()
            except BaseException as e:
                ### We want to pass any exception to the main thread
                ### but then there could be a delay in handling.
                ### What is a better solution?
                self.error("threaded_reader: {}",repr(e))
                self.error("threaded_reader: pass to the main thread")
                self._thread_exception = e
                ### set events so there is no waiting in the main thread
                ### main must now check and clear events
                self._packet_picked.set()
                self._nonempty.set()
            finally:
                if self._lock.locked(): self._lock.release()
            ### Slow it down for debugging. DON'T FORGET!!!
            # time.sleep(0.01)

    def read(self,keep=False,timeout=None):
        """Read the packet from the top of the queue.

        If timeout is not None, wait for at most timeout second.
        If meanwhile no packet arrived in the queue, return None.

        If keep is False, remove the packet from the queue, otherwise keep 
        it there.
        """
        ### Let's see if something bad happened in the thread.
        self._check_exception()
        self.debug( "read: queue size={}; queue_nonempty={}",
                    len(self._queue),
                    bool(self) )
        gotpack = self._nonempty.wait(timeout)
        ### Let's see if the thread got an exception while we were waiting
        self._check_exception()
        if gotpack:
            p = self._queue[0]
            self.debug( "read: got {} at {} from {} packets",
                          p.name,p.time,len(self))
            if not keep:
                self._lock.acquire()
                del self._queue[0]
                if not self._queue: self._nonempty.clear()
                self._lock.release()
            return p
        else: 
            self.debug("read: queue is empty, returning None after timeout")
            return None

    def pick(self,picker,until=None,timeout=0.5,keep=False):
        """Find and return all packets in the queue for which picker 
        evaluates to True and which arrived before the first packet for which
        until picker evaluates to true.
        Return with whatever found after timeout seconds, if the 'wait_for' packet 
        did not arrive. 
        Keep packets in the queue if keep is True, otherwise remove them.
        That does not includes the packet that marks the end of the search,
        unless it is also picked.
        """
        self._check_exception()
        if until is None:
            until = picker
        packs = list()
        indices = list()
        try:
            self._lock.acquire()
            if not any( map(until,self._queue) ):
                self.debug("pick: Didn'r reach until. Wait for the threaded reader")
                self._spack_picker = until
                self._spack_found.clear()
                self._lock.release()
                self._spack_found.wait(timeout)
                self._check_exception()
                self._lock.acquire()
            for i, p in enumerate(self._queue):
                if picker(p):
                    self.debug("pick: {}. picked a pack {}",i,p.name)
                    indices.append(i)
                    packs.append(p)
                if until(p):
                    self.debug("pick: reached until")
                    break
            self.debug("pick: found {} out of {} packs",
                       len(indices),len(self)    )
            if not keep:
                self.debug("pick: deleting found")
                for i in reversed(indices):
                    del self._queue[i]
            if not self._queue: self._nonempty.clear()
            return packs
        finally:
            self._spack_picker = None
            if self._lock.locked(): self._lock.release()
            
    def clear(self):
        "Clear the queue."
        self._lock.acquire()
        self._queue.clear()
        self._lock.release()

    def _check_exception(self):
        if self._thread_exception:
            e = self._thread_exception
            self._thread_exception = None
            self.debug("check_exception: {} detected in the thread",repr(e))
            ### thread sets events for the main thread to proceed
            ### we have to clear, if necessary
            self._packet_picked.clear()
            if not self._queue:
                self._nonempty.clear()
            raise e

    def _resync(self):
        """Seek the pipe to the start of the next packet"""
        found = False
        while not found:
            peek     = self._pipe.peek()
            position = peek.find(FVWM_PACK_START_b)
            if position == -1:
                self._pipe.read(len(peek))
                ### let's wait a bit
                time.sleep(0.05)
            else:
                self._pipe.read(position)
                found = True




import os as _os
import sys as _sys
import struct as _struct
import time as _time

from   .constants     import *
from   .packet_queue  import _packet_queue
### this is needed to set logging level
from   .packet        import packet
from   .exceptions    import *
from   .log           import _getloggers
from   .picker        import picker, glob, Glob

################################################################################
### Some helpers
def split_mask(mask):
    """Returns a list of all packet types matching the given mask
    """
    masks = list()
    cmask = 1
    while mask:
        if 1 & mask: masks.append(cmask)
        mask  >>= 1
        cmask <<= 1
    return masks

class _unique_id:
    """
    Instance of _unique_id is a callable object that returns a unique
    string with every invocation
    """
    
    uid = int(_time.perf_counter()*1000000000)
    def __call__(self,fmt="unique_id_0x{:x}"):
        self.uid += 1
        return fmt.format(self.uid)

unique_id =  _unique_id()

VERSION = "1.2.0"

################################################################################

class _fvwmvar:
    """
    Class for objects that provide access to FVWM variables.
    
    var.<name_of_var>

    returns value of the FVWM  variable in the same window context that
    the module was initiated.
    In the <name_of_var> underscores have to be used in place of dots 
    in the FVWM's names. The return value of the variable is always a 
    string. Attempts to assign to or delete FVWM variables result in
    IllegalOperation exception.

    var("<name_of_var1>","<name_of_var2>",...,context_window = None) 
    
    return a tuple of values of variables. In this invocation method it is 
    not necessary, (but is allowed) to replace dots with underscores in 
    variable names.
    context_window is id of the window in which context variable are to be 
    expanded. If None the the module's context window is assumed.
    If context_window is 0, then no context is assumed.
    E.g. provided that window with id equal cwid exists

    var("w.name","pointer.wx","pointer_wy",context_window=cwid) 

    will return the name of the window, and x- and y-coordinates of the 
    pointer within window (as a triple of strings).
    The return value is always a tuple (of strings), even if the number of 
    arguments is not greater then one.

    If a variable with the given name does not exist
    literal string '$[<name.of.var>]' is returned in either access
    methods. 

    The way this object works does not depend on the packet queue being empty.
    """

    _sep = "CrazySplitDelimiter3.1415"
    def __init__(self,module):
        super().__setattr__("_module",module)
        
    def __getattr__(self,var):
        vardots  = "$[{}]".format(var.replace("_","."))
        return self._module.getreply(vardots)

    def __setattr__(self,var,val):
        raise IllegalOperation("It is not possible to assign to Fvwm variables")

    def __delattr__(self,var):
        raise IllegalOperation("It is not possible to delete Fvwm variables")

    def __call__(self, *args,context_window=None):
        vardots = ["$[{}]".format(x.replace("_",".")) for x in args]
        varline = self._sep.join(vardots)
        reply = self._module.getreply(varline, context_window=context_window)
        values= reply.split(self._sep)
        if len(values) != len(args):
            raise FvwmError("fvwmvar: Something is wrong, "+
                            "more answers then questions or vice versa.",
                            values,args)
        return(values)
        
class _infostore:
    """An instance of _infostore class provide access to FVWM's
    infostore database.

    infostore.<name_of_var>
    infostore.<name_of_var> = val
    del infostore.<name_of_var>

    returns value, assigns and deletes the FVWM infostore variable.
    In the <name_of_var> underscores have to be used in place of dots 
    in the FVWM's names. The return value of the variable is always a 
    string.

    infostore("<name_of_var1>","<name_of_var2>",...) 
    
    return a tuple of values of variables. In this invocation method it is 
    necessary, (but is allowed) to replace dots with underscores in variable 
    names.
    The return value is always a tuple (of strings), even if the number of 
    arguments is not greater then one.

    If a variable with the given anme is absent from the FVWM's infostore 
    database, literal string '$[infostore.<name.of.var>]' is returned in 
    either access method. 

    The way this object works does not depend on the packet queue being empty.
    """
    
    _sep = "|CrazySplitDelimiter3.1415|"
    def __init__(self,module):
        super().__setattr__("_module",module)
        
    def __getattr__(self,var):
        vardots  = "$[infostore.{}]".format(var.replace("_","."))
        return self._module.getreply(vardots)

    def __setattr__(self,var,val):
        vardots  = var.replace("_",".")
        self._module.sendmessage("InfoStoreAdd {} ' {}'".
                                 format(vardots, str(val)))

    def __delattr__(self,var):
        vardots  = var.replace("_",".")
        self._module.sendmessage("InfoStoreRemove {}".format(vardots))

    def __call__(self, *args,context_window=None):
        vardots = [ "$[infostore.{}]".format(x.replace("_","."))
                    for x in args ]
        varline = self._sep.join(vardots)
        reply = self._module.getreply(varline).split(self._sep)
        if len(reply) != len(args):
            raise FvwmError("infostore: Something is wrong, "+
                            "more answers then questions or vice versa.",
                            str(reply),str(args))
        return(reply)
 
class _window(dict):
    """Instances of _window class are dictionaries
    containing key:value pairs corresponding to the fields in 
    M_CONFIGURE_WINDOW packet and other window packets, see
    fvwmpy.constants.M_FOR_WINLIST constant.
    Besides they contain additional values sent by fvwm during
    reply to Send_WindowList message. It has an additional method
    w.flag(int: i), that gives the value of i^th flag, see 
    file vpacket.h in fvwm source tree for the meaning of the flags.
    The __str__ method gives some human readable representation of the 
    instance of the _window class.

    ToDo: Access flags by meaningfull names.
    """

    def __getattr__(self,attr):
        return self[attr]

    def __setattr__(self,attr,val):
        self[attr] = val

    def flag(self,i):
        """Get the i^th flag of the window as a 0/1 integer
        """
        b,s = divmod(i,8)
        return self.flags[b] & ( 1 << s )

    def __str__(self):
        res = list()
        res.append("Fvwm Window: 0x{:x}".format(self["window"]) )
        for k,v in self.items():
            if k in {"window","frame"}:
                res.append("  | {} = 0x{:x}".format(k,v))
            else:
                res.append("  | {} = {}".format(k,v))
        return "\n".join(res)
    
    
class _winlist(dict):
    """Dictionary of all windows indexed by window id's.  

    Additional method .filter(conditions) gives an
    iterator over windows satisfying conditions, where conditions is a 
    string of conditions acceptable to FVWM's conditional commands and 
    have the same meaning.
    """
    def __init__(self,module):
        super().__init__()
        super().__setattr__("_module",module)
                
    def filter(self,conditions):
        """Return an iterator cycling through windows satisfying
        conditions.

        conditions  -- a multi-line string containing condition as in 
        FVWM's conditional commands.
        """
        self._module.push_masks(M_STRING|M_ERROR,0,0)
        cl = conditions.splitlines()
        cl = map(lambda x: x.strip(" \t,"),cl)
        cl = filter(None, cl)
        cond = ",".join(cl)
        self._module.debug( " Use condition {}",cond)
        self._module.sendmessage(
            "All ({}) SendToModule {} $[w.id]".
            format(cond, self._module.alias), context_window = 0 )
        self._module.sendmessage(
            "SendToModule {} finishedfilterwindows".
            format(self._module.alias), context_window = 0  )
        filteredlist = list()
        try:
            p = self._module.packets.read()
            self._module.debug( "Got {}",p.string )
            while p.ptype & M_STRING and p.string.startswith("0x"):
                filteredlist.append( int(p.string,0) )
                p = self._module.packets.read()
                self._module.debug( "Got {}",p.string )
            if p.ptype & M_ERROR:
                self._module.error("winlist: {}",p.string)
                raise FvwmError(p.string)
        finally:
            self._module.restore_masks()
                   
        for wid in filteredlist:
            yield self[wid]

    def __str__(self):
        res=list()
        for w in self.values():
            res.append(str(w))
        return "\n\n".join(res)

class _config(list):
    """Objects of this class contain configuration information
    supplied by FVWM in reply to Send_ConfigInfo command
    
    It is a list of module configuration lines.

    Additional attributes are:
        c.DesktopSize
        c.ImagePath
        c.XineramaConfig
        c.ClickTime
        c.IgnoreModifiers
        c.colorsets

    c.colorsets is a tuple if lists. Each list correspond to a colorset
    """

    _max_colorsets = int("0x40",16)
    def __init__(self):
        self.DesktopSize  = (None, None)
        self.ImagePath = tuple()
        self.XineramaConfig = tuple()
        self.ClickTime = None
        self.IgnoreModifiers = tuple()
        self.colorsets = [None for i in range(self._max_colorsets)]

    def __str__(self):
        res = list()
        res.append("FVWM configuartion database")
        res.append("  DesktopSize = {}".format(self.DesktopSize))
        res.append("  ImagePath = {}".format(self.ImagePath))
        res.append("  XineramaConfig = {}".format(self.XineramaConfig))
        res.append("  ClickTime = {}".format(self.ClickTime))
        res.append("  IgnoreModifiers = {}".format(self.IgnoreModifiers))
        res.append("  colorsets:")
        for i, c in enumerate(self.colorsets):
            res.append("    {:x} = {}".format(i,c))
        res.append("  module(s) configuration:")
        for cl in self:
            res.append("    "+cl)
        return "\n".join(res)
        
class fvwmpy:
    """Base class for developing Fvwm modules"""
   
    def __init__(self):
        self.me     = _os.path.split(_sys.argv[0])[1]
        if len(_sys.argv) < 6:
            raise FvwmLaunch("{}: Should only be executed by fvwm!".
                             format(self.alias))
        try:
            self._tofvwm     = _os.fdopen(int(_sys.argv[1]), "wb")
            self._fromfvwm   = _os.fdopen(int(_sys.argv[2]), "rb")
        except:
            raise FvwmLaunch("{}: Can not open read/write pipes".
                             format(self.alias))
        self.context_window = int(_sys.argv[4],0)
        self.context_deco   = int(_sys.argv[5],0);
        if _sys.argv[6:]:
            if not _sys.argv[6].startswith('-'):
                self._alias = _sys.argv[6]
                self.args =  _sys.argv[7:]
            elif _sys.argv[6] == '-':
                self.args =  _sys.argv[7:]
            else:
                self.args =  _sys.argv[6:]
        else:
            self.args = list()

        ( self.logger, self.debug, self.info,
          self.warn,   self.error, self.critical  ) = _getloggers(self.alias)
        self.logger.setLevel(L_WARN)

        self.handlers     = { pack : [] for pack in packetnames }
        ### We have to do that because mask.setter assumes 
        ### that _mask already exists.
        self._mask        = -1
        self._syncmask    = -1
        self._nograbmask  = -1
        self.mask         = 0
        self.syncmask     = 0
        self.nograbmask   = 0
        self._mask_stack  = list()
        self.winlist      = _winlist(self)
        self.config       = _config()
        self.var          = _fvwmvar(self)
        self.infostore    = _infostore(self)
        self.packets      = _packet_queue(self)
        
    @property
    def alias(self):
        try:
            return self._alias
        except AttributeError:
            return self.me

    def sendmessage_hook(self,msg,context_window,finished):
        """This method is called by m.sendmessage() method
        and is passed the same arguments.
        
        It doesn't do anything, but can be overloaded in a 
        derived class.
        """
        pass
    
    def sendmessage(self,msg, context_window=None, finished=False):
        """Send a possibly multiline string to fvwm using in context 
        window_context.
        If context_window==None, use context window in which module was 
        invoked. If context_window==0, use no context.
        If finished, notify FVWM that the module has finished working and 
        will exit soon.
        """
        if context_window is None:
            context_window = self.context_window
        lines = map( lambda x: x.strip(), msg.splitlines() )
        lines = filter(None,lines)
        lines = tuple(map(lambda l: l.encode(FVWM_STR_CODEX), lines))
        cw   = _struct.pack("L",context_window)
        for l in lines :
            self._tofvwm.write( b''.join( (cw,
                                           _struct.pack("L",len(l)),
                                           l,
                                           NOT_FINISHED) ) )
            self.debug(" Send message {}",l)
        if finished :
            self._tofvwm.write(b''.join( (cw,
                                          _struct.pack("L",3),
                                          b'NOP',
                                          FINISHED) ) )
        self._tofvwm.flush()
        self.sendmessage_hook(msg, context_window, finished)
        
    def finishedstartup(self):
        """
        Notify FVWM, that the module has finished configuring itself
        and is ready to work
        """
        
        self.debug("FINISHED STARTUP")
        self.sendmessage("NOP FINISHED STARTUP")
        
    def exit(self,n=0):
        """Exit from the module with exit status n"""
        
        self.unlock(finished=True)
        self._tofvwm.close()
        self._fromfvwm.close()
        self.info(" Exit")
        _sys.exit(n)

    def unlock(self,finished=False):
        """
        Send "NOP UNLOCK" command to FVWM.
        If finished=True, tell FVWM that the module will exit soon
        """
        
        self.sendmessage("NOP UNLOCK",finished)

    def mask_setter_hook(self, mask_type, m):
        """
        This is executed whenever m.mask, m.syncmask or m.grabmask 
        are set to new values. 
        
        mask_type is one of 'mask', 'syncmask' or 'nograbmask'.
        m is the new value of the mask. It does nothing,
        but can be overloaded in derived classes.
        """
        pass
        
    @property
    def mask(self):
        """Set or get mask. FVWM is notified of new value.
        """
        return self._mask
 
    @mask.setter
    def mask(self,m):
        """Set or get mask. FVWM is notified of new value
        """
        if self._mask == m: return
        self._mask = m
        ml = self._mask & ( M_EXTENDED_MSG - 1 )
        mu = (self._mask >> 32) | M_EXTENDED_MSG
        self.sendmessage("SET_MASK {}\nSET_MASK {}".format(ml,mu))
        self.mask_setter_hook("mask",m)

    @property
    def syncmask(self):
        """Set or get syncmask. FVWM is notified of new value.
        """

        return self._syncmask

    @syncmask.setter
    def syncmask(self,m):
        if self._syncmask == m: return
        self._syncmask = m 
        ml = self._syncmask & ( M_EXTENDED_MSG - 1 )
        mu = (self._syncmask >> 32) | M_EXTENDED_MSG
        self.sendmessage("SET_SYNC_MASK {}\nSET_SYNC_MASK {}".format(ml,mu))
        self.mask_setter_hook("syncmask",m)
        
    @property
    def nograbmask(self):
        """Set or get nograbmask. FVWM is notified of new value.
        """
        return self._nograbmask

    @nograbmask.setter
    def nograbmask(self,m):
        if self._nograbmask == m: return
        self._nograbmask = m
        ml = self._nograbmask & ( M_EXTENDED_MSG - 1 )
        mu = (self._nograbmask >> 32) | M_EXTENDED_MSG
        self.sendmessage( "SET_NOGRAB_MASK {}\nSET_NOGRAB_MASK {}".
                          format(ml,mu))
        self.mask_setter_hook("syncmask",m)

    def push_masks(self,mask,syncmask,nograbmask):
        "Temporarily assign new values to masks"
        if mask       is None: mask       = self.mask
        if syncmask   is None: syncmask   = self.syncmask
        if nograbmask is None: nograbmask = self.nograbmask
        self._mask_stack.append( (self.mask,self.syncmask,self.nograbmask) )
        self.mask, self.syncmask, self.nograbmask = (
            mask, syncmask, nograbmask ) 
        
    def restore_masks(self):
        "Restore previous values of masks"
        try:
           self.mask, self.syncmask, self.nograbmask = self._mask_stack.pop() 
        except IndexError:
            raise IllegalOperation(
                "Can not restore masks. Mask stack is empty" )

    def getreply(self,msg,context_window=None,timeout=0.5):
        """Send a string to FVWM in context context_window
        get the reply and return it
        """
        if context_window is None:
            context_window = self.context_window
        uid = unique_id()
        self.push_masks(self.mask|MX_REPLY,0,0)
        try:
            self.sendmessage( "Send_Reply {}{}".format(uid,msg),
                              context_window = context_window)
            replypicker = picker(mask = MX_REPLY,string =Glob(uid+"*") )
            packs = self.packets.pick(picker=picker(mask = MX_REPLY,
                                                  string =Glob(uid+"*") ),
                                    timeout=timeout)
        finally:
            self.restore_masks()
        if not packs:
            self.warn( "getreply: didn't get any reply from FVWM, return None")
            return
        else:
            return packs[0].string.replace(uid,"")

    def getconfig(self, handler=None, match=None, timeout=0.5):
        """Ask FVWM for module configuration information matching
        string in parameter match ('*'+m.alias if match==None).
        Pass the reply packets to handler (h_saveconfig if handler==None)
        """

        if match is None: match   = "*" + self.alias
        ### Ask first
        self.push_masks(self.mask|M_FOR_CONFIG,0,0)
        try:
            self.sendmessage("Send_ConfigInfo {}".format(match))
            if handler is None:
                handler = self.h_saveconfig
                self.info("getconfig: standard handler")
                self.config    = _config()
                self.rawconfig = list()
            packs = self.packets.pick( picker = picker(mask=M_FOR_CONFIG),
                                       until  = picker(mask=M_END_CONFIG_INFO),
                                       timeout = timeout )
            self.info( "getconfig: got {} config packets",len(packs))
        finally:
            self.restore_masks()
        for p in packs:
            handler(p)
        return packs[-1].ptype == M_END_CONFIG_INFO
            
    def getwinlist(self, handler = None, timeout=0.5):
        """Ask FVWM for the list of all windows.
        Pass replies to handler (h_updatewl if handler==None)
        """
        ### Ask FVWM first
        self.push_masks(self.mask|M_FOR_WINLIST,0,0)
        try:
            self.sendmessage("Send_WindowList")
            if handler is None:
                handler = self.h_updatewl
                self.winlist.clear()
            packs = self.packets.pick( picker = picker(mask = M_FOR_WINLIST),
                                       until  = picker(mask = M_END_WINDOWLIST),
                                       timeout = timeout )
            self.info( "getwinlist: got {} winlist packets",
                       len(packs))
        finally:
            self.restore_masks()
        for p in packs:
            handler(p)
        return packs[-1].ptype == M_END_WINDOWLIST
    
    def register_handler(self,mask,handler):
        """Add handler to the end of execution queues for all packets 
        matching mask.
        """
        for m in self.handlers:
            if (m & mask) and not (handler in self.handlers[m]):
                self.handlers[m].append(handler)

    def unregister_handler(self,mask,handler):
        """Remove handler from all execution queues for packets matching 
        masks. If handler is not in a queue, do nothing.
        """
        for m in self.handlers:
            if (m & mask):
                try:
                    self.handlers[m].remove(handler)
                except ValueError:
                    continue

    def call_handlers(self,p):
        """Execute all handlers in the queue for the packet p passing p as an 
        argument in the order they were registered.
        """
        for h in self.handlers[p.ptype]:
            h(p)
        
    def clear_handlers(self,mask):
        """Clear all queues for packets matching mask."""
        for h in self.handlers[p.ptype]:
            if h & mask:
                self.handlers[p.ptype] = list()
         
    def registered_handler(self,handler):
        """Return the mask matching all packet types for which handler 
        is registered.
        """
        mask = 0
        for m in self.handlers:
            if handler in self.handlers[m]:
                mask |= m
        return mask
    
            
    ### HANDLERS
    def h_saveconfig(self,p):
        """Handler. Packet types: M_FOR_CONFIG.

        This handler simply stores the information in the packet in the 
        config database.
        If packet p has wrong type IllegalOperation exception is raised.
        """
        if p.ptype == M_END_CONFIG_INFO: return
        if not p.ptype & M_FOR_CONFIG:
            raise IllegalOperation(
                "h_saveconfig: Packet must have type matching M_FOR_CONFIG" )
        ### FVWM is not consistent. Some strings have '\n' at the end.
        p.string = p.string.strip()
        ### For debugging DON'T FORGET!!!
        # self.rawconfig.append(p.string)
        
        if p.string == glob("[*]*"):
            self.config.append(p.string)
        elif p.string == glob("colorset *"):
            ### ToDo: parse colorsets
            cll = p.string.split()
            self.config.colorsets[int(cll[1],16)]=cll[2:]
        elif p.string == glob("desktopsize *"):
            cll = p.string.split()
            self.config.DesktopSize = ( int(cll[1]), int(cll[2]) )
        elif p.string == glob("imagepath *"):
            paths =  p.string.split()[1].strip().split(":")
            self.config.ImagePath   = tuple(paths)
        elif p.string == glob("xineramaconfig *"):
            cll = p.string.split()[1:]
            self.config.XineramaConfig = tuple(map(int,cll))
        elif p.string == glob("clicktime *"):
            self.config.ClickTime = int(p.string.split()[1])
        elif p.string == glob("ignoremodifiers *"):
            cll = p.string.split()[1:]
            self.config.IgnoreModifiers = tuple(map(int, cll))
        else:
            FvwmError("Can not parse the packet: {}".format(p.string))
            
    def h_unlock(self, p):
        """Handler. Packet types: M_ALL.

        This handler sends 'NOP UNLOCK' command to FVWM. It should be added to 
        the end of queues for packets in syncmask.
        """
        self.unlock()
        
    ### Update winlist
    def h_updatewl(self,p):
        """Handler. Packet types: M_FOR_WINLIST | M_DESTROY_WINDOW

        This handler updates the winlist database with the information in the 
        packet p.
        """
        if p.ptype & M_END_WINDOWLIST: return
        if not p.ptype & M_FOR_WINLIST | M_DESTROY_WINDOW:
            raise IllegalOperation(
                "h_updatewl: Packet must have type matching " +
                "M_FOR_WINLIST or M_DESTROY_WINDOW" )
        if p.ptype == M_DESTROY_WINDOW:
            try:               del self.winlist[p.window]
            except KeyError(): pass
        else:
            if p.window not in self.winlist:
                self.winlist[p.window] = _window()
            up = dict(p)
            for key in {"body","ptype","time"}:
                try:             del up[key]
                except KeyError: pass
            self.winlist[p.window].update(up)

    def h_exit(self,p):
        """Handler. Packet types: M_ALL.

        Exit the module.
        """
        self.exit()

    def h_nop(self,p):
        """Handler. Packet types: M_ALL.

        Do nothing.
        """
        pass

    h_pass = h_nop
    
    def run(self):
        """Mainloop.
        Read packets and execute corresponding handlers. 
        """
        self.debug(" Start main loop")
        while True:
            p = self.packets.read()
            self.call_handlers(p)
            


    



import fnmatch

################################################################################
### some helpers
from .log import _getloggers
from .constants import *
( logger, debug, info,
  warn,   error, critical  ) = _getloggers('fvwmpy:picker')
logger.setLevel(L_WARN)
################################################################################

class picker:
    def __init__(self, fcn=None, mask=None, **kwargs):
        if fcn is None:
            self.fcn = self._picker_factory(mask,**kwargs)
        elif mask is None and not kwargs:
            self.fcn = fcn
        else:
            raise ValueError( "Wrong arguments signature. "+
                              "Must be either picker(fcn=fcn) or " +
                              "picker(mask=mask,**kwargs)" )
        
    def __call__(self,p):
        return self.fcn(p)
    
    def __and__(self,other):
        def fcn(p):
            return self.fcn(p) and other.fcn(p)
        return picker(fcn=fcn)

    __rand__ = __and__
    
    def __or__(self,other):
        def fcn(p):
            return self.fcn(p) or other.fcn(p)
        return picker(fcn=fcn)

    __ror__ = __or__
    
    def __invert__(self):
        def fcn(p):
            return not self.fcn(p)
        return picker(fcn=fcn)

    @classmethod
    def _picker_factory(cls, mask=None,**kwargs):
        def fcn(p):
            debug("picker_fcn: working")
            if ( (mask is not None) and
                 (not p["ptype"] & mask) ):
                debug("Check {}=p['ptype'] ?= {}",
                      bin(p.get("ptype")),bin(mask))
                return False
            for k,v in kwargs.items():
                debug("Check {}=p[{}] ?= {}",p.get(k),k,v)
                if not ( k in p and p[k] == v ): return False
            return True
        return fcn

class glob(str):
    """Instance of glob is a string considered as a glob pattern.
    It may contain '*' and '?' characters and '[chars]' substrings.
    Pattern '[chars]' may also contain ranges like '[a-z0-9]'.
    '*' matches any substring, '?' matches any single character. 
    '[chars]' matches any sigle character which is one of 'chars'. 
    To match '*', '?', literally, use the corresponding character
    enclosed in '[]'. Check the documentation of the fnmatch package
    for more details.

    When glob instance is compared to another string with '==' or '!=' 
    operator, it checks whether another string matches the glob pattern. 
    Matching is case insensitive.

    E.g. "abc*efg?xyz" == glob("*c[*]EFg[?]x??") returns true.
    """

    def __str__(self):
        return "glob('{}')".format(super().__str__())
    
    __repr__ = __str__

    def __eq__(self,other):
        return fnmatch.fnmatchcase(other.lower(),self.lower())
    
    def __ne__(self,other):
        return not self.__eq__(other)

class Glob(str):
    """Instance of glob is a string considered as a glob pattern.
    It may contain '*' and '?' characters and '[chars]' substrings.
    Patterns '[chars]' may also contain ranges like '[a-z]'.
    '*' matches any substring, '?' matches any single character.
    '[chars]' matches any sigle character which is one of 'chars'. 
    To match '*', '?', literally, use the corresponding character
    enclosed in '[]'.

    When glob instance is compared to another string with '==' or '!=' 
    operator, it checks whether another string matches the glob pattern. 
    Matching is case sensitive.

    E.g. "abc*Efg?xyz" == Glob("*c[*]Efg[?]x??") returns true.
    """

    def __str__(self):
        return "Glob('{}')".format(super().__str__())
    
    __repr__ = __str__

    def __eq__(self,other):
        return fnmatch.fnmatchcase(str.__str__(other),str.__str__(self))
    
    def __ne__(self,other):
        return not self.__eq__(other)



def f():
    try:
        print("main")
        # raise Exception("In try")
    except:
        print("except")
        raise Exception("In except")
    else:
        print("Else block")
        raise Exception("In else")
    finally:
        print("finally")
        raise Exception("In finally")

class A:
    def __init__(self,a):
        self.a=a

    def __bool__(self):
        return self.a == 10

a=A(8)
def f():
    try:
        a=A(3)
        return a
    finally:
        a.a=5

        
